/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include <cstddef>
#include <fggl/data/procedural.hpp>
#include <fggl/data/model.hpp>

#include <glm/ext/matrix_transform.hpp>
#include <glm/gtc/quaternion.hpp>

#include <iostream>
#include <array>

#include <glm/geometric.hpp>

using namespace fggl::data;

// from https://www.khronos.org/opengl/wiki/Calculating_a_Surface_Normal
static glm::vec3 calcSurfaceNormal(glm::vec3 vert1, glm::vec3 vert2, glm::vec3 vert3) {
	const glm::vec3 edge1 = vert2 - vert1;
	const glm::vec3 edge2 = vert3 - vert1;
	return glm::normalize( glm::cross( edge1, edge2 ) );
}

static void computeNormalsDirect( fggl::data::Mesh& mesh, const fggl::data::Mesh::IndexType * colIdx, int nPoints) {

	// we're assuming all the normals are zero...
	for ( int i=0; i<nPoints; i++ ) {
		auto& vertex = mesh.vertex( colIdx[i] );
		vertex.normal = glm::vec3(0.0F);
	}

	// We're assuming each vertex appears only once (because we're not indexed)
	for (int i=0; i<nPoints; i += 3) {
		auto& v1 = mesh.vertex( colIdx[ i ] );
		auto& v2 = mesh.vertex( colIdx[ i + 1 ] );
		auto& v3 = mesh.vertex( colIdx[ i + 2 ] );

		const glm::vec3 normal = glm::normalize( calcSurfaceNormal( v1.posititon, v2.posititon, v3.posititon ) );
		v1.normal = normal;
		v2.normal = normal;
		v3.normal = normal;
	}
}

static void compute_normals(fggl::data::Mesh& mesh,
							const std::vector<Mesh::IndexType>& idxList, // source index
							const std::vector<Mesh::IndexType>& idxMapping // source-to-mesh lookup
							) {

	// clear the normals, so the summation below works correctly
	for (auto vertexIndex : idxMapping) {
		auto& vertex = mesh.vertex( vertexIndex );
		vertex.normal = ILLEGAL_NORMAL;
	}

	// we need to calculate the contribution for each vertex
	// this assumes IDXList describes a raw triangle list (ie, not quads and not a strip)
	for ( std::size_t i=0; i < idxList.size(); i += 3) {
		auto& v1 = mesh.vertex( idxMapping[idxList[i]] );
		auto& v2 = mesh.vertex( idxMapping[idxList[i + 1]] );
		auto& v3 = mesh.vertex( idxMapping[idxList[i + 2]] );

		// calculate the normal and area (formula for area the math textbook)
		float area = glm::length(glm::cross(v3.posititon - v2.posititon, v1.posititon - v3.posititon)) / 2;
		auto faceNormal = calcSurfaceNormal(v1.posititon, v2.posititon, v3.posititon);

		// weight the normal according to the area of the surface (bigger area = more impact)
		v1.normal += area * faceNormal;
		v2.normal += area * faceNormal;
		v3.normal += area * faceNormal;
	}

	// re-normalise the normals
	for (unsigned int vertexIndex : idxMapping) {
		auto& vertex = mesh.vertex( vertexIndex );
		vertex.normal = glm::normalize(vertex.normal);
	}
}

static void populateMesh(fggl::data::Mesh& mesh, const fggl::math::mat4 transform,
						 const int nIdx, const fggl::math::vec3* pos, const Mesh::IndexType* idx ) {

	auto* colIdx = new fggl::data::Mesh::IndexType[nIdx];

	// generate mesh
	for (int i=0; i<nIdx; i++) {
		glm::vec3 rawPos = transform * glm::vec4( pos[ idx[i] ], 1.0 );
		colIdx[ i ] = mesh.pushVertex( Vertex::from_pos(rawPos) );
		mesh.pushIndex( colIdx[i] );
	}

	computeNormalsDirect( mesh, colIdx, nIdx );

	delete[] colIdx;
}

static void populateMesh(fggl::data::Mesh& mesh,
						 const fggl::math::mat4 transform,
						 const std::vector<fggl::math::vec3>& posList,
						 const std::vector<fggl::data::Mesh::IndexType>& idxList) {

	// tmp store the resulting mesh indexes (incase the mesh has multiple primatives)
	std::vector<unsigned int> colIdx;
	colIdx.reserve(posList.size());

	// clion this thinks this loop is infinite, my assumption is it's gone bananas
	for (std::size_t i = 0; i < posList.size(); ++i) {
		glm::vec3 position = transform * fggl::math::vec4(posList[i], 1.0F);
		colIdx[i] = mesh.pushVertex( Vertex::from_pos(position) );
	}

	// use the remapped indexes for the mesh
	for (auto idx : idxList) {
		mesh.pushIndex( colIdx[idx] );
	}

	compute_normals(mesh, idxList, colIdx);
}

namespace fggl::data {

	static void quads2Tris(std::vector<Mesh::IndexType>& indexList, uint32_t stacks, uint32_t slices) {
		const auto HORZ_SIZE = slices + 1;

		for (uint32_t vertical = 0; vertical < stacks; vertical++) {
			for (uint32_t horz = 0; horz < slices; horz++) {
				const auto topLeft = vertical * HORZ_SIZE + horz;
				const auto topRight = vertical * HORZ_SIZE + (horz + 1);
				const auto bottomLeft = (vertical + 1) * HORZ_SIZE + horz;
				const auto bottomRight = (vertical + 1) * HORZ_SIZE + (horz + 1);

				indexList.push_back(topLeft);
				indexList.push_back(bottomLeft);
				indexList.push_back(topRight);

				indexList.push_back(topRight);
				indexList.push_back(bottomLeft);
				indexList.push_back(bottomRight);
			}
		}
	}

	void make_sphere(Mesh &mesh, const math::mat4& offset, uint32_t slices, uint32_t stacks) {

		std::vector<math::vec3> positions;

		float verticalAngularStride = math::PI / (float)stacks;
		float horizontalAngularStride = math::TAU / (float)slices;

		// calculate vertex positions
		for (uint32_t vertical = 0; vertical < (stacks + 1); vertical++) {
			float theta = (math::HALF_PI) - verticalAngularStride * (float)vertical;
			for ( uint32_t horz = 0; horz < (slices + 1); horz++ ) {
				float phi = horizontalAngularStride * (float)horz;
				math::vec3 position {
					cosf(theta) * cosf(phi),
					cosf(theta) * sinf(phi),
					sinf(theta)
				};
				positions.push_back(position);
			}
		}

		// combine the vertices into triangles
		std::vector<Mesh::IndexType> indexList;
		quads2Tris(indexList, stacks, slices);

		populateMesh(mesh, offset, positions, indexList);
	}


} // namespace fggl::data


fggl::data::Mesh fggl::data::make_triangle() {
    constexpr fggl::math::vec3 pos[]{
            {-0.5F, -0.5F, 0.0F},
            {0.5F,  -0.5F, 0.0F},
            {0.0F,  0.5F,  0.0F}
    };

    // add points
    fggl::data::Mesh mesh;
    for (auto po : pos) {
        mesh.push( Vertex::from_pos(po) );
    }

    // mesh
    for (int i = 0; i < 3; ++i) {
        mesh.pushIndex(i);
    }

    return mesh;
}

fggl::data::Mesh fggl::data::make_quad_xy() {
	constexpr fggl::math::vec3 pos[] {
		{-0.5F, -0.5F, 0.0F},
		{ 0.5F, -0.5F, 0.0F},
		{ 0.5F,  0.5F, 0.0F},
		{-0.5F,  0.5F, 0.0F}
	};
	constexpr int idx[] {
		0, 1, 3,
		3, 1, 2
	};

	fggl::data::Mesh mesh;
	std::array<int,4> colIdx{};
	for (int i = 0; i < 4; ++i){
		colIdx[ i ] = mesh.pushVertex( Vertex::from_pos(pos[i]));
	}

	for( auto i : idx ) {
		mesh.pushIndex( colIdx[ i ] );
	}

	return mesh;
}

fggl::data::Mesh fggl::data::make_quad_xz() {
	constexpr std::array<fggl::math::vec3, 4> pos {{
													   {-0.5F, 0.0F, -0.5F},
													   {0.5F, 0.0F, -0.5F},
													   {0.5F, 0.0F, 0.5F},
													   {-0.5F, 0.0F, 0.5F}
												   }};
	constexpr std::array<int, 6> idx {{
										  0, 1, 3,
										  3, 1, 2
									  }};

	fggl::data::Mesh mesh;
	std::array<int, 4> colIdx{};
	for (int i = 0; i < 4; ++i){
		colIdx[ i ] = mesh.pushVertex( Vertex::from_pos(pos[i]) );
	}

	for( auto i : idx ) {
		mesh.pushIndex(colIdx.at( i ));
	}

	return mesh;
}




fggl::data::Mesh fggl::data::make_cube(fggl::data::Mesh& mesh, const fggl::math::mat4& transform) {
	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	constexpr std::array<fggl::math::vec3, 8> pos {{
													   {-0.5, 0.5, -0.5}, // 0 TOP LOOP
													   {0.5, 0.5, -0.5}, // 1
													   {0.5, 0.5, 0.5}, // 2
													   {-0.5, 0.5, 0.5}, // 3
													   {-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
													   {0.5, -0.5, -0.5}, // 5
													   {0.5, -0.5, 0.5}, // 6
													   {-0.5, -0.5, 0.5}  // 7
												   }};

	constexpr std::array<Mesh::IndexType, 36> idx {{
														 0, 3, 1, // top
														 3, 2, 1,
														 0, 1, 4, // side 0 - 1
														 5, 4, 1,
														 1, 2, 5, // side 1 - 2
														 2, 6, 5,
														 3, 7, 2, // side 2 - 3
														 2, 7, 6,
														 0, 4, 3, // side 3 - 0
														 4, 7, 3,
														 4, 5, 7, // bottom
														 7, 5, 6,
													 }};
	populateMesh(mesh, transform, idx.size(), pos.data(), idx.data());
	return mesh;
}

fggl::data::Mesh fggl::data::make_slope(fggl::data::Mesh& mesh, const fggl::math::mat4& transform) {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	// FIXME remove 2 and 3 and renumber the index list accordingly
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr Mesh::IndexType idx[] {
		0, 7, 1, // ramp
		7, 6, 1,
		0, 1, 4, // side 0 - 1
		5, 4, 1,
		1, 6, 5, // ramp side 1
		4, 7, 0, // ramp side 2
		4, 5, 7, // bottom
		7, 5, 6,
	};

	int nIdx = sizeof(idx) / sizeof(int);
	populateMesh(mesh, transform, nIdx, pos, idx);
	return mesh;
}


fggl::data::Mesh fggl::data::make_ditch(fggl::data::Mesh& mesh, const fggl::math::mat4& transform) {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	// FIXME remove 2 and renumber the index list accordingly
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr Mesh::IndexType idx[] {
		0, 3, 1, // top
		3, 6, 1,
		0, 1, 4, // side 0 - 1
		5, 4, 1,
		1, 6, 5,
		3, 7, 6,
		0, 4, 3, // side 3 - 0
		4, 7, 3,
		4, 5, 7, // bottom
		7, 5, 6,
	};

	int nIdx = sizeof(idx) / sizeof(int);
	populateMesh(mesh, transform, nIdx, pos, idx);
	return mesh;
}

fggl::data::Mesh fggl::data::make_point(fggl::data::Mesh& mesh, const fggl::math::mat4& transform) {

	// done as two loops, top loop is 0,1,2,3, bottom loop is 4,5,6,7
	constexpr fggl::math::vec3 pos[] {
		{-0.5,  0.5, -0.5}, // 0 TOP LOOP
		{ 0.5,  0.5, -0.5}, // 1
		{ 0.5,  0.5,  0.5}, // 2
		{-0.5,  0.5,  0.5}, // 3
		{-0.5, -0.5, -0.5}, // 4 BOTTOM LOOP
		{ 0.5, -0.5, -0.5}, // 5
		{ 0.5, -0.5,  0.5}, // 6
		{-0.5, -0.5,  0.5}  // 7
	};

	constexpr Mesh::IndexType idx[] {
		0, 7, 5, // top
		7, 6, 5,
		0, 5, 4, // side 0 - 1
		0, 4, 7, // side 3 - 0
		4, 5, 7, // bottom
		7, 5, 6,
	};

	int nIdx = sizeof(idx) / sizeof(int);
	populateMesh(mesh, transform, nIdx, pos, idx);
	return mesh;
}



/*
constexpr float cubeVertex[8 * 3] = {
	0, 0, 0,
	0, 1, 0,
	1, 1, 0,
	1, 0, 0,
	0, 0, 1,
	0, 1, 1,
	1, 1, 1,
	0, 0, 0,
};

constexpr int cubeIndexes[6 * 4] = {
	7, 6, 5, 4,
	0, 1, 2, 3,
	6, 7, 3, 2,
	5, 6, 2, 1,
	4, 5, 1, 0,
	7, 4, 0, 3,
};


void quadsToTris(fggl::data::Mesh& data, const std::vector<unsigned int>& idx, const int* quads, const int nQuads) {
	for (int p = 0; p < nQuads; p++) {
		int offset = p * 4;
		// first tri
		data.pushIndex( idx[ quads[offset + 0 ] ] );
		data.pushIndex( idx[ quads[offset + 1 ] ] );
		data.pushIndex( idx[ quads[offset + 2 ] ] );

		// second tri
		data.pushIndex( idx[ quads[offset + 2 ] ] );
		data.pushIndex( idx[ quads[offset + 3 ] ] );
		data.pushIndex( idx[ quads[offset + 0 ] ] );
	}
}

inline glm::vec3 calcNormal(Vertex prev, Vertex next, Vertex curr) {
	return glm::cross( prev.posititon - curr.posititon, next.posititon - curr.posititon );
}

void computeNormals(Mesh& mesh) {

	const int idxes = mesh.indexCount();
	const int vertCount = mesh.vertexCount();

	// zero the normals
	for (int i=0; i<vertCount; i++) {
		Vertex& vert = mesh.vertex( i );
		vert.normal = glm::vec3(0.0f);
	}

	// calculate normals for all triangles
	for (int tri=0; tri<idxes; tri += 3) {
		// get next 3 points
		Vertex& pa = mesh.vertex( tri );
		Vertex& pb = mesh.vertex( tri + 1 );
		Vertex& pc = mesh.vertex( tri + 2 );

		pa.normal += calcNormal( pb, pc, pa );
		pb.normal += calcNormal( pc, pa, pb );
		pc.normal += calcNormal( pa, pb, pc );
	}

	// normalise
	for (int i=0; i<vertCount; i++) {
		Vertex& vert = mesh.vertex( i );
		vert.normal = glm::normalize( vert.normal );
	}

}

Mesh fggl::data::make_cube() {

	int nVerts = sizeof(cubeVertex) / sizeof(cubeVertex[0]) / 3;
	int nQuads = sizeof(cubeIndexes) / sizeof(cubeIndexes[0]) / 4;

	std::vector< unsigned int > vertIndex;

	Mesh cube;

	// stick the vertex data into the mesh
	for ( int i=0; i < nVerts; i++ ) {
		fggl::data::Vertex vert;
		vert.posititon = fggl::math::vec3( cubeVertex[i], cubeVertex[i + 1], cubeVertex[i+2] );

		auto idx = cube.pushVertex(vert);
		vertIndex.push_back(idx);
	}

	// triagulate the quads
	quadsToTris(cube, vertIndex, cubeIndexes, nQuads);
	compute_normals(cube);

	return cube;
}*/

