/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 17/04/22.
//

#include <fggl/gui/widget.hpp>

namespace fggl::gui {

	void buttonBorder( gfx::Path2D& path, glm::vec2 pos, glm::vec2 size ) {
		// outer box
		path.colour( {1.0f, 0.0f, 0.0f} );
		path.pathTo( { pos.x + size.x, pos.y } );
		path.pathTo( { pos.x + size.x, pos.y + size.y } );
		path.pathTo( { pos.x, pos.y + size.y } );
		path.close();

		// inner box
		math::vec2 innerTop { pos.x + 5, pos.y + 5 };
		math::vec2 innerBottom { pos.x + size.x - 5, pos.y + size.y - 5 };

		path.colour( {1.0f, 1.0f, 0.0f} );
		path.moveTo( { innerTop.x, innerTop.y } );
		path.pathTo( { innerBottom.x, innerTop.y } );
		path.pathTo( { innerBottom.x, innerBottom.y } );
		path.pathTo( { innerTop.x, innerBottom.y } );
		path.pathTo( { innerTop.x, innerTop.y } );
	}

	void draw_box( gfx::Path2D& path, glm::vec2 topLeft, glm::vec2 bottomRight ) {
		path.moveTo( { topLeft.x, topLeft.y } );
		path.pathTo( { bottomRight.x, topLeft.y } );
		path.pathTo( { bottomRight.x, bottomRight.y } );
		path.pathTo( { topLeft.x, bottomRight.y } );
		path.pathTo( { topLeft.x, topLeft.y } );
	}

	void draw_progress( gfx::Path2D& path, glm::vec2 topLeft, glm::vec2 size, float value ) {
		const auto bottomRight { topLeft + size };

		// background
		path.colour( {0.5f, 0.5f, 0.5f} );
		draw_box( path, topLeft, bottomRight );

		// fill
		math::vec2 innerTop { topLeft.x + 5, topLeft.y + 5 };
		math::vec2 innerBottom { bottomRight.x - 5, bottomRight.y - 5 };

		// figure out how wide the bar should be
		float barWidth = (innerBottom.x - innerTop.x) * value;
		float trueBottom = innerBottom.x;
		innerBottom.x = innerTop.x + barWidth;

		// draw the bar
		path.colour( {0.8f, 0.0f, 0.0f} );
		draw_box( path, innerTop, innerBottom );

		// part of the bar that's not filled in
		math::vec2 emptyTop { innerBottom.x, innerTop.y };
		math::vec2 emptyBottom { trueBottom, innerBottom.y };
		path.colour( {0.4f, 0.0f, 0.0f} );
		draw_box( path, emptyTop, emptyBottom );
	}

	void draw_slider( gfx::Path2D& path, glm::vec2 topLeft, glm::vec2 size, float value ) {
		draw_progress( path, topLeft, size, value );

		// dimensions
		const auto bottomRight { topLeft + size };
		const math::vec2 innerTop { topLeft.x + 5, topLeft.y + 5 };
		const math::vec2 innerBottom { bottomRight.x - 5, bottomRight.y - 5 };

		// selector bar
		float trackWidth = innerBottom.x - innerTop.x;
		float selectorValue = trackWidth * value;
		float selectorWidth = 6;

		math::vec2 selectorTop { innerTop.x + selectorValue - ( selectorWidth/2), topLeft.y };
		math::vec2 selectorBottom { selectorTop.x + selectorWidth, bottomRight.y };
		path.colour( {1.0f, 1.0f, 1.0f} );
		draw_box( path, selectorTop, selectorBottom );
	}

	void draw_button( gfx::Path2D& path, glm::vec2 pos, glm::vec2 size, bool active, bool pressed) {
		// locations
		math::vec2 outerTop { pos };
		math::vec2 outerBottom { pos + size };
		math::vec2 innerTop { pos.x + 5, pos.y + 5 };
		math::vec2 innerBottom { pos.x + size.x - 5, pos.y + size.y - 5 };

		math::vec3 baseColour{ 0.5f, 0.5f, 0.5f };

		if ( active ) {
			baseColour *= 1.2f;
		}

		if ( pressed ) {
			baseColour *= 0.8f;
		}

		math::vec3 lightColour{ baseColour * 1.2f };
		math::vec3 darkColour{ baseColour * 0.8f };
		if ( pressed ) {
			// flip light and dark for selected buttons
			auto tmp = darkColour;
			darkColour = lightColour;
			lightColour = tmp;
		}

		// bottom side
		path.colour( lightColour );
		path.moveTo( outerTop );
		path.pathTo( innerTop );
		path.pathTo( { innerBottom.x, innerTop.y } );
		path.pathTo( { outerBottom.x, outerTop.y } );
		path.pathTo( outerTop );

		// left side
		path.colour( lightColour );
		path.moveTo( outerTop );
		path.pathTo( innerTop );
		path.pathTo( { innerTop.x, innerBottom.y } );
		path.pathTo( { outerTop.x, outerBottom.y } );
		path.pathTo( outerTop );

		// top side
		path.colour( darkColour );
		path.moveTo( { outerTop.x, outerBottom.y} );
		path.pathTo( { innerTop.x, innerBottom.y} );
		path.pathTo( innerBottom );
		path.pathTo( outerBottom );
		path.pathTo( { outerTop.x, outerBottom.y}  );

		// right side
		path.colour( darkColour );
		path.moveTo( outerBottom );
		path.pathTo( innerBottom );
		path.pathTo( { innerBottom.x, innerTop.y } );
		path.pathTo( { outerBottom.x, outerTop.y } );
		path.pathTo( outerBottom  );

		// inner box
		path.colour( baseColour );
		draw_box( path, innerTop, innerBottom );
	}
}