/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 27/06/22.
//

#ifndef FGGL_ASSETS_MANAGER_HPP
#define FGGL_ASSETS_MANAGER_HPP

#include <string_view>
#include <map>
#include <functional>
#include <memory>

#include "fggl/assets/types.hpp"
#include "fggl/util/safety.hpp"

namespace fggl::assets {

	class AssetManager {
		public:
			constexpr const static modules::ModuleService service = modules::make_service("fggl::assets::Manager");
			using AssetGUID = std::string;

			AssetManager() = default;
			virtual ~AssetManager() = default;

			// no move, no copy.
			AssetManager(const AssetManager&) = delete;
			AssetManager& operator=(const AssetManager&) = delete;
			AssetManager(AssetManager&&) = delete;
			AssetManager& operator=(AssetManager&&) = delete;

			inline AssetRefRaw getRaw(const AssetGUID& guid) const {
				return m_registry.at(guid).asset;
			}

			template<typename T>
			AssetRef<T> get(const AssetGUID& guid) const {
				return std::dynamic_pointer_cast<T>(getRaw(guid));
			}

			inline void require(const AssetGUID& guid) {
				m_registry.at(guid).refCount++;
			}

			inline void release(const AssetGUID& guid) {
				m_registry.at(guid).refCount--;
			}

		private:
			struct AssetRecord {
				std::shared_ptr<Asset> asset = nullptr;
				std::size_t refCount;

				inline ~AssetRecord(){
					if ( asset != nullptr ) {
						asset->release();
					}
				}
			};
			std::map<AssetGUID, AssetRecord> m_registry;
	};


} // namespace fggl::assets

#endif //FGGL_ASSETS_MANAGER_HPP
