/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 01/05/22.
//

#ifndef FGGL_AUDIO_OPENAL_AUDIO_HPP
#define FGGL_AUDIO_OPENAL_AUDIO_HPP

#include <AL/al.h>
#include <AL/alc.h>

#include "fggl/audio/audio.hpp"
#include "fggl/data/storage.hpp"
#include "fggl/math/types.hpp"

#include <string>
#include <iostream>
#include <memory>

namespace fggl::audio::openal {

	enum class AudioType {
		MONO_8 = AL_FORMAT_MONO8,
		MONO_16 = AL_FORMAT_MONO16,
		STEREO_8 = AL_FORMAT_STEREO8,
		STEREO_16 = AL_FORMAT_STEREO16
	};

	static void checkError(std::string context) {
		auto code = alGetError();
		if ( code == AL_NO_ERROR) {
			return;
		}

		// now we check the error message
		std::string error = "unknown";
		switch ( code ) {
			case ALC_INVALID_DEVICE:
				error = "Invalid Device";
				break;
			case ALC_INVALID_CONTEXT:
				error = "Invalid Context";
				break;
			case ALC_INVALID_ENUM:
				error = "Invalid enum";
				break;
			case ALC_INVALID_VALUE:
				error = "Invalid value";
				break;
			case ALC_OUT_OF_MEMORY:
				error = "Out of memory";
				break;
			default:
				error = "unknown error";
		}

		std::cerr << "OpenAL error: " << context << ": " << error << std::endl;
	}

	class AudioBuffer {
		public:
			AudioBuffer() : m_buffer(0) {
				alGenBuffers(1, &m_buffer);
			}
			~AudioBuffer() {
				alDeleteBuffers(1, &m_buffer);
			}

			inline void setData(AudioType type, void* data, ALsizei size, ALsizei frequency) {
				alBufferData(m_buffer, (ALenum)type, data, size, frequency);
			}

			inline ALuint value() {
				return m_buffer;
			}
		private:
			ALuint m_buffer;
	};


	class AudioSource {
		public:
			AudioSource() : m_source(0), m_splat() {
				alGenSources(1, &m_source);
			}
			~AudioSource(){
				alDeleteSources(1, &m_source);
			}

			inline void play() {
				alSourcePlay( m_source );
			}

			inline void stop() {
				alSourceStop(m_source);
			}

			inline void pause() {
				alSourcePause( m_source );
			}

			inline void rewind() {
				alSourceRewind( m_source );
			}

			inline void play(AudioBuffer& buffer, bool looping) {
				alSourcei(m_source, AL_BUFFER, buffer.value());
				alSourcei( m_source, AL_LOOPING, looping ? AL_TRUE : AL_FALSE);
			}

			inline void play(AudioClip& clip, bool looping) {
				checkError("pre play");

				AudioType format = clip.channels == 1 ? AudioType::MONO_8 : AudioType::STEREO_8;
				m_splat.setData(format, clip.data, clip.sampleCount, clip.sampleRate);
				checkError("saving to buffer");

				alSourcei(m_source, AL_BUFFER, m_splat.value());
				alSourcei( m_source, AL_LOOPING, looping ? AL_TRUE : AL_FALSE);
				checkError("setting parameters");

				play();
			}

			inline void velocity(math::vec3& value) {
				alSource3f(m_source, AL_VELOCITY, value.x, value.y, value.z);
			}
			inline void position(math::vec3& value) {
				alSource3f(m_source, AL_POSITION, value.x, value.y, value.z);
			}

			void direction(math::vec3& value) {
				alSource3f(m_source, AL_DIRECTION, value.x, value.y, value.z);
			}

		private:
			ALuint m_source;
			AudioBuffer m_splat;
	};


	class AudioServiceOAL : public AudioService {
		public:
			explicit AudioServiceOAL(data::Storage* storage) : m_device(alcOpenDevice(nullptr)), m_storage(storage) {
				if ( m_device != nullptr ) {
					m_context = alcCreateContext(m_device, nullptr);
					alcMakeContextCurrent(m_context);
					checkError("context setup");

					m_defaultSource = std::make_unique<AudioSource>();
					checkError("default source setup");
				}
			}

			~AudioServiceOAL() override {
				if ( m_device != nullptr ) {
					alcDestroyContext(m_context);
					alcCloseDevice(m_device);
				}
			}

			void play(const std::string& filename, bool looping = false) override;
			void play(AudioClip& clip, bool looping = false) override;

		private:
			ALCdevice* m_device;
			ALCcontext* m_context{nullptr};
			std::unique_ptr<AudioSource> m_defaultSource{nullptr};
			data::Storage* m_storage = nullptr;
	};


} // namespace fggl::audio::openal

namespace fggl::audio {
	using openAL = openal::AudioServiceOAL;

} // namepace fggl::audio

#endif //FGGL_AUDIO_OPENAL_AUDIO_HPP
