/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_GFX_OGL_SHADER_HPP
#define FGGL_GFX_OGL_SHADER_HPP

#include <cstdio>
#include <fggl/gfx/ogl/backend.hpp>
#include <fggl/data/storage.hpp>

#include <filesystem>
#include <stdexcept>
#include <unordered_map>

namespace fggl::gfx {

	constexpr GLuint INVALID_SHADER_ID = 0;

	struct ShaderConfig {
		// required
		std::string name;
		std::string vertex;
		std::string fragment;

		// optional parts
		std::string geometry;
		bool hasGeom = false;

		constexpr static ShaderConfig named(const std::string& name, bool hasGeom = false) {
			return {
				.name = name,
				.vertex = name + "_vert.glsl",
				.fragment = name + "_frag.glsl",
				.geometry = hasGeom ? name + "_geom.glsl" : "",
				.hasGeom = hasGeom
			};
		}
	};

	struct ShaderSources {
		std::string name;
		std::string vertexSource;
		std::string fragmentSource;
		std::string geometrySource;
	};

	struct BinaryCache {
		void *data = nullptr;
		GLsizei size = 0;
		GLenum format = GL_INVALID_ENUM;
	};

	class ShaderCache {
		public:
			ShaderCache(fggl::data::Storage* storage);
			~ShaderCache() = default;

			GLuint load(const ShaderConfig &config);
			GLuint load(const ShaderSources &sources, bool allowBinaryCache);

			GLuint getOrLoad(const ShaderConfig &config);

			GLuint get(const std::string &name);

			/**
			 * Fallback pipelines.
			 *
			 * All fallback pipelines are prefixed with "fallback_". They should be used as 'last resort' if there is
			 * no matching shader on disk for this game.
			 */
			void initFallbackPipelines();

		private:
			std::shared_ptr<fggl::data::Storage> m_storage;
			std::unordered_map<std::string, GLuint> m_shaders;

			// extensions
			void setupIncludes();

			// opengl operations
			bool readAndCompileShader(const std::string& filename, GLuint shader);
			bool compileShaderFromSource(const std::string& source, GLuint);

			// file io operations
			bool loadFromDisk(GLuint pid, const std::string &pipelineName);
			void saveToDisk(GLuint pid, const std::string &pipelineName);

			bool m_binary;
			void cacheSave(GLuint pid, BinaryCache *cache);
			bool cacheLoad(GLuint pid, const BinaryCache *cache);
	};

}

#endif
